<?php

declare(strict_types = 1);

namespace Drupal\unpublished_file\Plugin\UnpublishedFile\FileSource;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\file\FileInterface;
use Drupal\unpublished_file\FileSourcePluginBase;

/**
 * Detects file source relations from entity_usage reference records.
 *
 * This plugin returns the file source entities and the source entity files by
 * simply querying the entity_usage module 'entity_usage.usage' service.
 *
 * @FileSource(
 *   id = "entity_usage",
 *   label = @Translation("Entity Usage reference"),
 *   description = @Translation("Detects file source relations from entity_usage reference records."),
 * )
 */
class EntityUsageReference extends FileSourcePluginBase {

  /**
   * {@inheritdoc}
   */
  public function getAccessibleSourceEntities(FileInterface $file, AccountInterface $account): array {
    return $this->doGetAccessibleSourceEntities($file, $account);
  }

  /**
   * {@inheritdoc}
   */
  public function hasAccessibleSourceEntities(FileInterface $file, AccountInterface $account): bool {
    return !empty($this->doGetAccessibleSourceEntities($file, $account, TRUE));
  }

  /**
   * Returns a list of accessible source entities given a file and a user.
   *
   * @param \Drupal\file\FileInterface $file
   *   The file entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The user account being checked.
   * @param bool $returnOnFirstFound
   *   (optional) Whether to return on first publicly available source entity
   *   found. Used as an optimization when the caller wants only to check if
   *   there is at least one published source entity. Defaults to FALSE.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface[]
   *   A list of host/source entities.
   */
  protected function doGetAccessibleSourceEntities(FileInterface $file, AccountInterface $account, bool $returnOnFirstFound = FALSE): array {
    $sources = [];

    foreach ($this->entityUsage->listSources($file) as $sourceEntityTypeId => $usages) {
      $sourceEntityDefinition = $this->entityTypeManager->getDefinition($sourceEntityTypeId);
      $sourceEntityStorage = $this->entityTypeManager->getStorage($sourceEntityTypeId);
      $ids = $sourceEntityStorage->getQuery()
        ->accessCheck(FALSE)
        ->condition($sourceEntityDefinition->getKey('id'), array_keys($usages), 'IN')
        ->execute();
      if ($ids) {
        foreach ($sourceEntityStorage->loadMultiple($ids) as $sourceEntity) {
          if ($sourceEntity->access('view', $account)) {
            $sources[] = $sourceEntity;
            if ($returnOnFirstFound) {
              // Optimization.
              return $sources;
            }
          }
        }
      }
    }

    return $sources;
  }

  /**
   * {@inheritdoc}
   */
  public function getFilesForSource(ContentEntityInterface $source): array {
    $files = [];
    foreach ($this->entityUsage->listTargets($source, $source->getRevisionId()) as $targetEntityTypeId => $usage) {
      if ($this->isFileEntityType($targetEntityTypeId)) {
        $files[$targetEntityTypeId] = array_keys($usage);
      }
    }
    return $files;
  }

}
