<?php

declare(strict_types = 1);

namespace Drupal\unpublished_file;

use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AnonymousUserSession;
use Drupal\Core\StreamWrapper\StreamWrapperManager;
use Drupal\file\FileInterface;
use Drupal\file\FileRepositoryInterface;

/**
 * Default implementation for 'unpublished_file' service.
 */
class UnpublishedFile implements UnpublishedFileInterface {

  /**
   * The anonymous user.
   */
  protected AccountInterface $anonymous;

  /**
   * Constructs a new service instance.
   *
   * @param \Drupal\Component\Plugin\PluginManagerInterface $fileSourceManager
   *   The file source plugin manager service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   The file system service.
   * @param \Drupal\file\FileRepositoryInterface $fileRepository
   *   The file repository service.
   */
  public function __construct(
    protected PluginManagerInterface $fileSourceManager,
    protected ConfigFactoryInterface $configFactory,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected FileSystemInterface $fileSystem,
    protected FileRepositoryInterface $fileRepository,
  ) {
    $this->anonymous = new AnonymousUserSession();
  }

  /**
   * {@inheritdoc}
   */
  public function setSourceFilesVisibility(EntityInterface $source): void {
    /** @var \Drupal\Core\Entity\ContentEntityInterface $source */
    if (!$this->isSourceEntity($source)) {
      return;
    }

    /** @var \Drupal\Core\Entity\ContentEntityInterface|null $originalSource */
    $originalSource = $source->original;
    if ($originalSource && $originalSource->access('view', $this->anonymous) === $source->access('view', $this->anonymous)) {
      // The host entity is updating but the visibility for the anonymous user
      // remains unchanged. Nothing to do.
      return;
    }

    // Collect source files from plugins.
    $files = [];
    foreach ($this->fileSourceManager->getDefinitions() as $pluginId => $pluginDefinition) {
      /** @var \Drupal\unpublished_file\FileSourcePluginInterface $plugin */
      $plugin = $this->fileSourceManager->createInstance($pluginId);
      $files = array_merge_recursive($files, $plugin->getFilesForSource($source));
    }

    foreach ($files as $fileEntityTypeId => $fileIds) {
      // Different plugins may return same files.
      $fileIds = array_unique($fileIds);
      $fileStorage = $this->entityTypeManager->getStorage($fileEntityTypeId);
      /** @var \Drupal\file\FileInterface $file */
      foreach ($fileStorage->loadMultiple($fileIds) as $file) {
        $hasAccessibleSources = $this->hasAnonymousAccessibleUsage($file);
        // @todo What about `$settings['file_additional_public_schemes']`?
        if (!$hasAccessibleSources && str_starts_with($file->getFileUri(), 'public://')) {
          $this->moveFile($file, 'private://unpublished-file/' . StreamWrapperManager::getTarget($file->getFileUri()));
        }
        elseif ($hasAccessibleSources && str_starts_with($file->getFileUri(), 'private://unpublished-files/')) {
          [, $path] = explode('/', StreamWrapperManager::getTarget($file->getFileUri()), 2);
          $this->moveFile($file, "public://$path");
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getAccessibleSourceEntities(FileInterface $file, AccountInterface $account): array {
    $sources = [];

    // @todo Not sure excluding temporary files is correct.
    if (!$file->isPermanent()) {
      // Don't mess with temporary files.
      return $sources;
    }

    foreach ($this->fileSourceManager->getDefinitions() as $pluginId => $pluginDefinition) {
      /** @var \Drupal\unpublished_file\FileSourcePluginInterface $plugin */
      $plugin = $this->fileSourceManager->createInstance($pluginId);
      $sources += $plugin->getAccessibleSourceEntities($file, $account);
    }

    return $sources;
  }

  /**
   * Checks whether a file has at least one anonymous-accessible usage.
   *
   * @param \Drupal\file\FileInterface $file
   *   The file entity to be checked.
   *
   * @return bool
   *   Whether a file has at least one anonymous-accessible usage.
   */
  protected function hasAnonymousAccessibleUsage(FileInterface $file): bool {
    foreach ($this->fileSourceManager->getDefinitions() as $pluginId => $pluginDefinition) {
      /** @var \Drupal\unpublished_file\FileSourcePluginInterface $plugin */
      $plugin = $this->fileSourceManager->createInstance($pluginId);
      if ($plugin->hasAccessibleSourceEntities($file, $this->anonymous)) {
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Moves a given file to a given destination.
   *
   * @param \Drupal\file\FileInterface $file
   *   The file entity.
   * @param string $destinationUri
   *   The destination URI.
   */
  protected function moveFile(FileInterface $file, string $destinationUri): void {
    $destinationDir = $this->fileSystem->dirname($destinationUri);
    if (!$this->fileSystem->prepareDirectory($destinationDir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) {
      throw new \RuntimeException("Directory $destinationDir cannot be created or it's not writable.");
    }
    $this->fileRepository->move($file, $destinationUri);
  }

  /**
   * Checks whether a given entity acts as a usage source.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to be checked.
   *
   * @return bool
   *   Whether the entity acts as a usage source.
   */
  protected function isSourceEntity(EntityInterface $entity): bool {
    if ($entity instanceof ContentEntityInterface) {
      $sourceEntityTypes = (array) $this->configFactory
        ->get('entity_usage.settings')
        ->get('track_enabled_source_entity_types');
      return empty($sourceEntityTypes) || in_array($entity->getEntityTypeId(), $sourceEntityTypes, TRUE);
    }
    return FALSE;
  }

}
